/*
 *  linux/fs/qnx4fs/dir.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/fs.h>
#include <linux/qnx4fs_fs.h>

static int
qnx4fs_readdir (struct file *fp, void *dirent, filldir_t filldir)
{
	struct inode *dip = fp->f_dentry->d_inode;
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	unsigned long phys;
	long nblocks, block;
	int slot, namelen;
	ino_t ino;

	nblocks = qnx4fs_nblocks(dip);
	block = fp->f_pos >> QNX4FS_DIRENTS_PER_BLOCK_BITS;
	slot = fp->f_pos & (QNX4FS_DIRENTS_PER_BLOCK-1);
	for (; block < nblocks; block++, slot = 0) {
		phys = qnx4fs_map_block(dip, block);
		if (!phys)
			continue;
		bp = bread (dip->i_dev, phys - 1, QNX4FS_BLOCK_SIZE);
		if (bp == NULL)
			continue;
		qdp = (qnx4fs_dirent_t *)bp->b_data + slot;
		for (; slot < QNX4FS_DIRENTS_PER_BLOCK; slot++, qdp++) {
			if (qdp->i.status & QNX4FS_FILE_LINK) {
				namelen = QNX4FS_LONG_NAME_MAX;
				ino = link2ino(qdp);
			} else if (qdp->i.status & QNX4FS_FILE_USED) {
				namelen = QNX4FS_SHORT_NAME_MAX;
				ino = direct2ino(phys, slot);
			} else {
				continue;
			}
			if (ino == QNX4FS_INODES_INO
			    || ino == QNX4FS_SB(dip->i_sb)->s_bitmap_ino) {
				continue;
			}
			namelen = strnlen(qdp->i.name, namelen);
			if (filldir(dirent, qdp->i.name, namelen,
				    (block << QNX4FS_DIRENTS_PER_BLOCK_BITS) | slot,
				    ino, DT_UNKNOWN) < 0) {
				brelse(bp);
				goto out;
			}
		}
		brelse(bp);
	}
out:
	UPDATE_ATIME(dip);
	fp->f_pos = (block << QNX4FS_DIRENTS_PER_BLOCK_BITS) | slot;
	fp->f_version = dip->i_version;

	return 0;
}

struct file_operations
qnx4fs_dir_ops = {
	read:		generic_read_dir,
	readdir:	qnx4fs_readdir
};
