/*
 *  linux/fs/qnx4fs/file.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/pagemap.h>
#include <linux/fs.h>
#include <linux/qnx4fs_fs.h>

#if 0
unsigned long
qnx4fs_map_block (struct inode *ip, long block)
{
	struct buffer_head *bp;
	qnx4fs_xb_t *xb;
	qnx4fs_xt_t *xt;
	unsigned long phys;
	long xbnr, offset;
	int nxt;

	if (block < QNX4FS_I(ip)->i_xt0.size)
		return (QNX4FS_I(ip)->i_xt0.start + block);
	xbnr = QNX4FS_I(ip)->i_xbnr;
	offset = QNX4FS_I(ip)->i_xt0.size;
	while (xbnr) {
		bp = bread(ip->i_dev, xbnr - 1, QNX4FS_BLOCK_SIZE);
		if (bp == NULL)
			return 0;
		xb = (qnx4fs_xb_t *)bp->b_data;
		if (memcmp(xb->xb_signature, "IamXblk\0", 8)) {
			brelse(bp);
			return 0;
		}
		nxt = xb->xb_nxt;
		xt = xb->xb_xts;
		for (; nxt > 0; nxt--) {
			if (block < offset + xt->size) {
				phys = xt->start + block - offset;
				brelse(bp);
				return phys;
			}
			offset += xt->size;
			xt++;
		}
		xbnr = xb->xb_next;
		brelse (bp);
	}
	return 0;
}
#else
unsigned long
qnx4fs_map_block (struct inode *ip, long block)
{
	struct buffer_head *bp;
	qnx4fs_xb_t *xb;
	qnx4fs_xt_t *xt;
	unsigned long phys;
	long xbnr, offset;
	int xtnr, nxt;

	xbnr = QNX4FS_I(ip)->i_xbnr;
	xtnr = QNX4FS_I(ip)->i_xtnr;
	if (block >= QNX4FS_I(ip)->i_offset) {
		if (block < QNX4FS_I(ip)->i_offset_end) {
			return (QNX4FS_I(ip)->i_phys + block - QNX4FS_I(ip)->i_offset);
		} else for (offset = QNX4FS_I(ip)->i_offset_end;
			    xbnr;
			    xbnr = le32_to_cpu(xb->xb_next),
			    brelse(bp),
			    xtnr = -1) {
			bp = bread (ip->i_dev, xbnr - 1, QNX4FS_BLOCK_SIZE);
			if (bp == NULL)
				return 0;
			xb = (qnx4fs_xb_t *)bp->b_data;
			if (memcmp(xb->xb_signature, "IamXblk\0", 8)) {
				brelse (bp);
				return 0;
			}
			nxt = xb->xb_nxt;
			if (xtnr >= nxt - 1)
				continue;
			xt = &xb->xb_xts[++xtnr];
			for (; xtnr < nxt; xtnr++, xt++) {
				offset += le32_to_cpu(xt->size);
				if (block < offset) {
					phys = le32_to_cpu(xt->start)
					       + le32_to_cpu(xt->size)
					       + block - offset;
					QNX4FS_I(ip)->i_offset = offset - le32_to_cpu(xt->size);
					QNX4FS_I(ip)->i_offset_end = offset;
					goto found;
				}
			}
		};
	} else if (block < QNX4FS_I(ip)->i_xt0.size) {
		return (QNX4FS_I(ip)->i_xt0.start + block);
	} else for (offset = QNX4FS_I(ip)->i_offset;
		    xbnr;
		    xbnr = le32_to_cpu(xb->xb_prev),
		    brelse(bp),
		    xtnr = QNX4FS_MAX_XTS_PER_XBLK) {
		bp = bread(ip->i_dev, xbnr - 1, QNX4FS_BLOCK_SIZE);
		if (bp == NULL)
			return 0;
		xb = (qnx4fs_xb_t *)bp->b_data;
		if (memcmp(xb->xb_signature, "IamXblk\0", 8)) {
			brelse (bp);
			return 0;
		}
		if (xtnr <= 0)
			continue;
		nxt = xb->xb_nxt;
		xt = &xb->xb_xts[--xtnr];
		for (; xtnr >= 0; xtnr--, xt--) {
			offset -= le32_to_cpu(xt->size);
			if (block >= offset) {
				phys = le32_to_cpu(xt->start) + block - offset;
				QNX4FS_I(ip)->i_offset = offset;
				QNX4FS_I(ip)->i_offset_end = offset + le32_to_cpu(xt->size);
				goto found;
			}
		}
	};
	printk(KERN_ERR "qnx4fs: block %ld for ino %ld not found\n", block, ip->i_ino);
	return 0;
found:
	QNX4FS_I(ip)->i_phys = le32_to_cpu(xt->start);
	QNX4FS_I(ip)->i_xtnr = xtnr;
	QNX4FS_I(ip)->i_xbnr = xbnr;
	brelse(bp);

	return phys;
}
#endif

int
qnx4fs_get_block (struct inode *ip, long block,
		  struct buffer_head *bh_result, int create)
{
	unsigned long phys;
	long nblocks;

	nblocks = (QNX4FS_I(ip)->i_mmu_private + QNX4FS_BLOCK_SIZE-1)
		>> QNX4FS_BLOCK_SIZE_BITS;
	if (create && block >= nblocks) {
//		QNXDBG((": expand %d\n", (int)block));

		phys = qnx4fs_iexpand(ip);
		if (!phys)
			return -ENOSPC;
		bh_result->b_state |= (1UL << BH_New);
	} else {
		phys = qnx4fs_map_block (ip, block);
		if (!phys)
			return -EIO;
	}
	bh_result->b_dev = ip->i_dev;
	bh_result->b_blocknr = phys - 1;
	bh_result->b_state |= (1UL << BH_Mapped);
	return 0;
}

struct file_operations qnx4fs_file_ops = {
	llseek:		generic_file_llseek,
	read:		generic_file_read,
	write:		generic_file_write,
	mmap:		generic_file_mmap
};

struct inode_operations qnx4fs_file_inode_ops = {
	truncate:	qnx4fs_truncate
};

static int
qnx4fs_readpage (struct file *filp, struct page *pp)
{
	return block_read_full_page (pp, qnx4fs_get_block);
}

static int
qnx4fs_writepage (struct page *pp)
{
	return block_write_full_page(pp, qnx4fs_get_block);
}

static int
qnx4fs_prepare_write (struct file *filp, struct page *pp,
		      unsigned from, unsigned to)
{
	return cont_prepare_write(pp, from, to, qnx4fs_get_block,
				  &QNX4FS_I(pp->mapping->host)->i_mmu_private);
}

static int
qnx4fs_bmap (struct address_space *mapping, long block)
{
	return generic_block_bmap(mapping, block, qnx4fs_get_block);
}

struct address_space_operations qnx4fs_address_space_ops = {
	readpage:	qnx4fs_readpage,
	writepage:	qnx4fs_writepage,
	sync_page:	block_sync_page,
	prepare_write:	qnx4fs_prepare_write,
	commit_write:	generic_commit_write,
	bmap:		qnx4fs_bmap
};
